//
//  It is file 'tot_def.hpp'
//
//  Copyright (C) 1992, 1993 Serge Pachkovsky, ps@che.nsk.su
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 1, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
#ifdef GEN_HUGE_FUNCS
#ifdef NELSON_ESCAPE
HugeInline void
Totals::Fill( Context *c, Boolean IgnoreBoard, int )
{
#ifdef DEBUG
if( c->GetMaxIndex() + 1 > CharsCount ){
    cerr << "Totals::Fill sanity check: static dimension overflow\n" ;
    exit( -1 ) ;
    }
#endif
for(;;){
	int	    i	= c->GetMaxIndex() ;
	UINT16	    max = 0 ;
	UINT16	   *tp	= counts + i + 2 ;

    for( *tp = 0 ; i >= 0 ; i--, tp-- ){
	    CountType	cnt = c->GetCount( i ) ;

	tp[ -1 ] = tp[ 0 ] ;
	if( IgnoreBoard || ! scoreboard.Checked( c->IndexToChar( i ) ) )
	    tp[ -1 ] += cnt ;
	if( cnt > max )
	    max = cnt ;
	}
    counts[ 0 ] = counts[ 1 ] + 1 ;
    if( max > 0 )
	counts[ 0 ] += ( UINT16( CharsCount - c->GetMaxIndex() ) * c->GetMaxIndex() )
		       / ( ( CharsCount * max ) / ( (CountTypeMax+1)/CharsCount ) ) ;
    if( counts[ 0 ] < MaximumArifmeticScale )
	break ;
    #ifdef STATS
    Model2::active_model->TotalCountOverflows++ ;
    #endif
    c->RescaleCharCounts() ;
    }

    int 	i = c->GetMaxIndex() ;

for( ; i >= 0 ; i-- )
    if( c->GetCount( i ) != 0 )
	scoreboard.Set( c->IndexToChar( i ) ) ;

entries = c->GetMaxIndex() + 2 ;
}
#else
HugeInline void
Totals::Fill( Context *c, Boolean IgnoreBoard, int codes_count )
{
#ifdef DEBUG
if( c->GetMaxIndex() + 1 > CharsCount ){
    cerr << "Totals::Fill sanity check: static dimension overflow\n" ;
    exit( -1 ) ;
    }
#endif
for(;;){
	int	    i		  = c->GetMaxIndex() ;
	int	    local_chars   = 0 ;
	UINT16	   *tp		  = counts + i + 1 ;
	CountType   max_cnt	  = 0 ;

    for( tp[1] = 0 ; i >= 0 ; i--, tp-- ){
	    CountType	cnt = c->GetCount( i ) ;

	tp[ 0 ] = tp[ 1 ] ;
	if( cnt != 0 ){
	    if( ! scoreboard.Checked( c->IndexToChar( i ) ) ){
		local_chars++ ;
		tp[ 0 ] += cnt ;
		if( cnt > max_cnt )
		    max_cnt = cnt ;
		}
	    else
	    if( IgnoreBoard ){
		tp[ 0 ] += cnt ;
		if( cnt > max_cnt )
		    max_cnt = cnt ;
		}
	    }
	}
    counts[ 0 ]  = counts[ 1 ] + 1 ;

	int	    seen_chars	  = local_chars + scoreboard.CheckCount() ;
	int	    escaped_chars = codes_count - seen_chars ;

    if( escaped_chars < 0 ){
	#ifdef DEBUG
	if( c->context_level > 0 ){
	     /*
   	      *	It _should_ happen in context 0, so I do not care...
	      */
	     cerr << "Totals::Fill(): incorrect estimation of characters count\n" ;
	     cerr << "codes_count = " << codes_count << " seen_chars = " 
	          << seen_chars << " escaped_chars = " << escaped_chars << "\n" ;
	     c->DumpContextInfo() ;
       	     }
	#endif
	escaped_chars = 0 ;
	}
    if( max_cnt > 0 )
	counts[ 0 ] += ( seen_chars * escaped_chars ) / ( codes_count * max_cnt ) ;

    if( counts[ 0 ] < MaximumArifmeticScale )
	break ;
    #ifdef STATS
    Model2::active_model->TotalCountOverflows++ ;
    #endif
    c->RescaleCharCounts() ;
    }

    int 	i = c->GetMaxIndex() ;

for( ; i >= 0 ; i-- )
    if( c->GetCount( i ) != 0 )
	scoreboard.Set( c->IndexToChar( i ) ) ;

entries = c->GetMaxIndex() + 2 ;
}
#endif
#endif

inline UINT16
Totals::GetScale( void ) const
{
return counts[ 0 ] ;
}

inline UINT16
Totals::LowCount( int index ) const
{
return counts[ index + 2 ] ;
}

inline UINT16
Totals::HighCount( int index ) const
{
return counts[ index + 1 ] ;
}

inline void
Totals::ClearScores( void )
{
scoreboard.Clear() ;
}

inline int
Totals::SearchCount( UINT16 key ) const
{
    int 	low = 0, high = entries, middle ;

#ifdef DEBUG
if( key >= GetScale() ){
    cerr << "Totals::SearchCount sanity check: count is " << key << " while scale is " << GetScale() << "\n" ;
    exit( -1 ) ;
    }
#endif
while( high - low > 1 ){
    middle = ( low + high ) / 2 ;
    if( key >= counts[ middle ] )
	 high = middle ;
    else low  = middle ;
    }
//
//  In the case some character counts were zero, we need to
//  skip to the right edge of interval. We will stop sometime,
//  because counts[ entries ] is always zero.
//
while( counts[ low ] > key ) low++ ;
return low - 2 ;
}

